/*
 * Copyright (c) 2013-2015, Georgia Tech Research Corporation
 * All rights reserved.
 *
 * Author(s): Jeongseok Lee <jslee02@gmail.com>
 *
 * Georgia Tech Graphics Lab and Humanoid Robotics Lab
 *
 * Directed by Prof. C. Karen Liu and Prof. Mike Stilman
 * <karenliu@cc.gatech.edu> <mstilman@cc.gatech.edu>
 *
 * This file is provided under the following "BSD-style" License:
 *   Redistribution and use in source and binary forms, with or
 *   without modification, are permitted provided that the following
 *   conditions are met:
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
 *   CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
 *   INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 *   MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *   DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
 *   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 *   USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 *   AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *   LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *   ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *   POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef KIDO_DYNAMICS_SCREWJOINT_HPP_
#define KIDO_DYNAMICS_SCREWJOINT_HPP_

#include <string>

#include <Eigen/Dense>

#include "kido/dynamics/SingleDofJoint.hpp"

namespace kido {
namespace dynamics {

/// class ScrewJoint
class ScrewJoint : public SingleDofJoint
{
public:

  friend class Skeleton;

  struct UniqueProperties
  {
    /// Rotational axis
    Eigen::Vector3d mAxis;

    /// Translational pitch
    double mPitch;

    UniqueProperties(const Eigen::Vector3d& _axis = Eigen::Vector3d::UnitZ(),
                     double _pitch = 0.1);

    virtual ~UniqueProperties() = default;
  };

  struct Properties : SingleDofJoint::Properties,
                      ScrewJoint::UniqueProperties
  {
    Properties(
        const SingleDofJoint::Properties& _singleDofProperties =
                                              SingleDofJoint::Properties(),
        const ScrewJoint::UniqueProperties& _screwProperties =
                                              ScrewJoint::UniqueProperties());
    virtual ~Properties() = default;
  };

  ScrewJoint(const ScrewJoint&) = delete;

  /// Destructor
  virtual ~ScrewJoint();

  /// Set the Properties of this ScrewJoint
  void setProperties(const Properties& _properties);

  /// Set the Properties of this ScrewJoint
  void setProperties(const UniqueProperties& _properties);

  /// Get the Properties of this ScrewJoint
  Properties getScrewJointProperties() const;

  /// Copy the Properties of another ScrewJoint
  void copy(const ScrewJoint& _otherJoint);

  /// Copy the Properties of another ScrewJoint
  void copy(const ScrewJoint* _otherJoint);

  /// Copy the Properties of another ScrewJoint
  ScrewJoint& operator=(const ScrewJoint& _otherJoint);

  // Documentation inherited
  virtual const std::string& getType() const override;

  /// Get joint type for this class
  static const std::string& getStaticType();

  // Documentation inherited
  virtual bool isCyclic(size_t _index) const override;

  ///
  void setAxis(const Eigen::Vector3d& _axis);

  ///
  const Eigen::Vector3d& getAxis() const;

  ///
  void setPitch(double _pitch);

  ///
  double getPitch() const;

protected:

  /// Constructor called by Skeleton class
  ScrewJoint(const Properties& _properties);

  // Documentation inherited
  virtual Joint* clone() const override;

  // Documentation inherited
  virtual void updateLocalTransform() const override;

  // Documentation inherited
  virtual void updateLocalJacobian(bool _mandatory=true) const override;

  // Documentation inherited
  virtual void updateLocalJacobianTimeDeriv() const override;

protected:

  /// ScrewJoint Properties
  UniqueProperties mScrewP;

public:
  // To get byte-aligned Eigen vectors
  EIGEN_MAKE_ALIGNED_OPERATOR_NEW
};

}  // namespace dynamics
}  // namespace kido

#endif  // KIDO_DYNAMICS_SCREWJOINT_HPP_
